unit BoolProp;

// Demonstrate Delphi 5's new owner-drawn property editors.
// These editos show a check box for Boolean properties
// and for set elements.
//
// Copyright  1999 Tempest Software, Inc.

interface

uses Windows, SysUtils, Classes, Graphics, TypInfo, DsgnIntf;

type
  TSetElementPropertyEx = class(TSetElementProperty)
  public
    procedure PropDrawValue(Canvas: TCanvas; const Rect: TRect;
      Selected: Boolean); override;
  end;

  TSetPropertyEx = class(TSetProperty)
  private
    fEnumInfo: PTypeInfo; // type info for the enumerated type
  protected
    property EnumInfo: PTypeInfo read fEnumInfo;
  public
    procedure Initialize; override;
    function GetAttributes: TPropertyAttributes; override;
    procedure GetProperties(Proc: TGetPropEditProc); override;
    procedure GetValues(Proc: TGetStrProc); override;
    procedure SetValue(const Value: string); override;
    procedure ListDrawValue(const Value: string; Canvas: TCanvas;
      const Rect: TRect; Selected: Boolean); override;
    procedure ListMeasureHeight(const Value: string; Canvas: TCanvas;
      var Height: Integer); override;
    procedure ListMeasureWidth(const Value: string; Canvas: TCanvas;
      var Width: Integer); override;
  end;

  TBoolPropertyEx = class(TBoolProperty)
  public
    procedure PropDrawValue(Canvas: TCanvas; const Rect: TRect;
      Selected: Boolean); override;
  end;
  TBooleanPropertyEx = class(TEnumProperty)
  public
    procedure PropDrawValue(Canvas: TCanvas; const Rect: TRect;
      Selected: Boolean); override;
  end;

procedure Register;

implementation

uses Controls, Grids, Menus, Forms;

var
  Checked: TBitmap;
  Unchecked: TBitmap;

// Load the standard Windows checkbox bitmaps. The Obm_CheckBoxes bitmap
// is a grid of 12 different styles. The two in the upper-left corner
// are the traditional checkbox bitmaps.
procedure Initialize;
var
  Bitmap: TBitmap;
  Src, Dst: TRect;
begin
  Checked := TBitmap.Create;
  Unchecked := TBitmap.Create;

  Bitmap := TBitmap.Create;
  try
    Bitmap.Handle := LoadImage(0, PChar(Obm_CheckBoxes), Image_Bitmap, 0, 0,
        Lr_DefaultSize or Lr_CreateDibSection or Lr_Shared or Lr_LoadMap3DColors);
    if Bitmap.Handle = 0 then
      RaiseLastWin32Error;
    Src.Left := 0;
    Src.Top := 0;
    Src.Right := Bitmap.Width div 4;
    Src.Bottom := Bitmap.Height div 3;
    Unchecked.Height := Src.Bottom - Src.Top - 1;
    Unchecked.Width := Src.Right - Src.Left - 1;
    Dst := Src;
    Unchecked.Canvas.CopyRect(Dst, Bitmap.Canvas, Src);

    Checked.Height := Unchecked.Height;
    Checked.Width := Unchecked.Width;
    Src.Left := Src.Right;
    Src.Right := Src.Left + Bitmap.Width div 4;
    Checked.Canvas.CopyRect(Dst, Bitmap.Canvas, Src);
  finally
    Bitmap.Free;
  end;
end;

procedure Register;
begin
  RegisterPropertyEditor(TypeInfo(Boolean), nil, '', TBooleanPropertyEx);
  RegisterPropertyEditor(TypeInfo(ByteBool), nil, '', TBoolPropertyEx);
  RegisterPropertyEditor(TypeInfo(WordBool), nil, '', TBoolPropertyEx);
  RegisterPropertyEditor(TypeInfo(LongBool), nil, '', TBoolPropertyEx);
end;

// Published properties are limited to 32 elements.
type
  TSet = set of 0..31;

{ TSetPropertyEx }

function TSetPropertyEx.GetAttributes: TPropertyAttributes;
begin
  Result := inherited GetAttributes + [paValueList, paSortList];
end;

procedure TSetPropertyEx.GetProperties(Proc: TGetPropEditProc);
var
  I: 0..31;
  Data: PTypeData;
begin
  Data := GetTypeData(EnumInfo);
  for I := Data.MinValue to Data.MaxValue do
    Proc(TSetElementPropertyEx.Create(Self, I));
end;

procedure TSetPropertyEx.GetValues(Proc: TGetStrProc);
var
  I: 0..31;
  Data: PTypeData;
begin
  Data := GetTypeData(EnumInfo);        // type data of the enumerated type
  for I := Data.MinValue to Data.MaxValue do
    Proc(GetEnumName(EnumInfo, I));
end;

procedure TSetPropertyEx.Initialize;
begin
  inherited;
  fEnumInfo := GetTypeData(GetPropType).CompType^;
end;

// Draw an item in the drop-down list. Display the checked or
// unchecked box for each item.
procedure TSetPropertyEx.ListDrawValue(const Value: string;
  Canvas: TCanvas; const Rect: TRect; Selected: Boolean);
var
  IsChecked: Boolean;
  OrdValue: Integer;
begin
  OrdValue := GetOrdValue;
  IsChecked := GetEnumValue(EnumInfo, Value) in TSet(OrdValue);
  Canvas.FillRect(Rect);
  Canvas.TextRect(Rect, Rect.Left + Checked.Width + 2, Rect.Top + 1, Value);
  if IsChecked then
    Canvas.Draw(Rect.Left + 1, Rect.Top + 1, Checked)
  else
    Canvas.Draw(Rect.Left + 1, Rect.Top + 1, Unchecked);
end;

procedure TSetPropertyEx.ListMeasureHeight(const Value: string;
  Canvas: TCanvas; var Height: Integer);
begin
  if Height < Checked.Height then
    Height := Checked.Height;
end;

procedure TSetPropertyEx.ListMeasureWidth(const Value: string;
  Canvas: TCanvas; var Width: Integer);
begin
  Width := Width + Checked.Width + 2;
end;

// When the user picks an item from the drop-down list,
// toggle that element of the set.
procedure TSetPropertyEx.SetValue(const Value: string);
var
  OrdValue: Integer;
  EnumValue: 0..31;
begin
  OrdValue := GetOrdValue;
  EnumValue := GetEnumValue(EnumInfo, Value);
  if EnumValue in TSet(OrdValue) then
    Exclude(TSet(OrdValue), EnumValue)
  else
    Include(TSet(OrdValue), EnumValue);
  SetOrdValue(OrdValue);
end;

// Draw a check box and the Boolean text label (i.e., True or False).
procedure DrawBoolCheckBox(Canvas: TCanvas; const Rect: TRect; const Value: string);
begin
  Canvas.FillRect(Rect);
  Canvas.TextRect(Rect, Rect.Left + Checked.Width + 2, Rect.Top + 1, Value);
  if Value = BooleanIdents[False] then
    Canvas.Draw(Rect.Left + 1, Rect.Top + 1, UnChecked)
  else
    Canvas.Draw(Rect.Left + 1, Rect.Top + 1, Checked);
end;

{ TSetElementPropertyEx }
// Display a checkbox for each item, showing whether or not is
// is a member of the set.
procedure TSetElementPropertyEx.PropDrawValue(Canvas: TCanvas;
  const Rect: TRect; Selected: Boolean);
begin
  DrawBoolCheckBox(Canvas, Rect, Value);
end;

{ TBoolPropertyEx }
// Display a check box for ByteBool, WordBool, and LongBool items.
// The user cannot click to check or uncheck, but double-click works.
procedure TBoolPropertyEx.PropDrawValue(Canvas: TCanvas; const Rect: TRect;
  Selected: Boolean);
begin
  DrawBoolCheckBox(Canvas, Rect, Value);
end;

{ TBooleanPropertyEx }
// Display a check box for Boolean items.
// The user cannot click to check or uncheck, but double-click works.
procedure TBooleanPropertyEx.PropDrawValue(Canvas: TCanvas;
  const Rect: TRect; Selected: Boolean);
begin
  DrawBoolCheckBox(Canvas, Rect, Value);
end;


initialization
  // Get the standard checkbox bitmaps.
  Initialize;
finalization
  FreeAndNil(Checked);
  FreeAndNil(Unchecked);
end.
