unit ColorProperty;

// Color property editor that has subproperties for the
// red, green, and blue elements of a color. This property
// editor demonstrates how you can use the TNestedProperty
// editor to create your own subproperties for any property.
//
// Copyright  1999 Tempest Software, Inc.

interface

uses Windows, SysUtils, Classes, Graphics, DsgnIntf;

type
  // Define a new type for RGB values, so you can easily see
  // which values are meant to be TColor (which might be a system
  // color) and which are exclusively RGB triplets.
  TRGB = type DWORD;
  // An RGB triplet contains three TColorElement values.
  TColorElement = type Byte;

  // Parent property editor. Register TColorSubproperty for any
  // property of type TColor. It's just like a TColorProperty editor,
  // except that it has subproperties for the Red, Green, and Blue
  // elements of a color.
  TColorSubproperty = class(TColorProperty)
  public
    function GetAttributes: TPropertyAttributes; override;
    procedure GetProperties(Proc: TGetPropEditProc); override;
  end;

  // Abstract base class for each color element.
  TColorElementProperty = class(TNestedProperty)
  protected
    // Call GetOrdValue and convert the TColor value to a RGB value.
    function GetRGBValue: TRGB;
    function GetRGBValueAt(Index: Integer): TRGB;

    function GetColorValue(Color: TRGB): TColorElement; virtual; abstract;
    function SetColorValue(Value: TColorElement; Color: TRGB): TRGB; virtual; abstract;
  public
    function AllEqual: Boolean; override;
    function GetValue: string; override;
    procedure SetValue(const Value: string); override;
  end;

  // Each of red, green, and blue have their own concrete derivations
  // from TColorElement. The TColorSubproperty editor creates
  // instances of each color element subproperty editor.
  TRedElementProperty = class(TColorElementProperty)
  public
    function GetName: string; override;
    function GetColorValue(Color: TRGB): TColorElement; override;
    function SetColorValue(Value: TColorElement; Color: TRGB): TRGB; override;
  end;
  TGreenElementProperty = class(TColorElementProperty)
  public
    function GetName: string; override;
    function GetColorValue(Color: TRGB): TColorElement; override;
    function SetColorValue(Value: TColorElement; Color: TRGB): TRGB; override;
  end;
  TBlueElementProperty = class(TColorElementProperty)
  public
    function GetName: string; override;
    function GetColorValue(Color: TRGB): TColorElement; override;
    function SetColorValue(Value: TColorElement; Color: TRGB): TRGB; override;
  end;

procedure Register;

implementation

// Subproperty names.
resourcestring
  sRed = 'Red';
  sGreen = 'Green';
  sBlue = 'Blue';

// You need to register only the parent property editor. This editor
// replaces the standard TColorProperty editor.
procedure Register;
begin
  RegisterPropertyEditor(TypeInfo(TColor), nil, '', TColorSubproperty);
end;

{ TColorSubproperty }
// TColorSubproperty is just like TColorProperty, except that it
// has subproperties.
function TColorSubproperty.GetAttributes: TPropertyAttributes;
begin
  Result := inherited GetAttributes + [paSubproperties];
end;

// Create each subproperty editor and pass it to Proc.
// Delphi frees the instances when it is done using them.
procedure TColorSubproperty.GetProperties(Proc: TGetPropEditProc);
begin
  Proc(TRedElementProperty.Create(Self));
  Proc(TGreenElementProperty.Create(Self));
  Proc(TBlueElementProperty.Create(Self));
end;


{ TColorElement }
// Return True is all selected components have the same
// value for the color element.
function TColorElementProperty.AllEqual: Boolean;
var
  I: Integer;
  Compare: TColorElement;
begin
  if PropCount > 1 then
  begin
    // Get the color element of the first selected component.
    Compare := GetColorValue(GetRGBValue);
    for I := 1 to PropCount - 1 do
      // Compare the first with the rest of the components.
      if Compare <> GetColorValue(GetRGBValueAt(I)) then
      begin
        // If any one is different, return False.
        Result := False;
        Exit;
      end;
  end;
  // They're all the same.
  Result := True;
end;

// Call GetOrdValue and convert the TColor value to a RGB value.
function TColorElementProperty.GetRGBValue: TRGB;
begin
  Result := ColorToRGB(GetOrdValue);
end;

// Return a string representation of the color element value.
function TColorElementProperty.GetValue: string;
begin
  Result := Format('$%2.2x', [GetColorValue(GetRGBValue)]);
end;

// Call GetOrdValueAt and convert the TColor value to a RGB value.
function TColorElementProperty.GetRGBValueAt(Index: Integer): TRGB;
begin
  Result := ColorToRGB(GetOrdValueAt(Index));
end;

// Set the parent color when the color element changes to the
// given value. If "Value" is invalid, the Object Inspector
// will catch the exception that StrToInt raises.
procedure TColorElementProperty.SetValue(const Value: string);
begin
  SetOrdValue(SetColorValue(StrToInt(Value), GetRGBValue));
end;


{ TRedElement }
// Return the red part of a color.
function TRedElementProperty.GetColorValue(Color: TRGB): TColorElement;
begin
  Result := GetRValue(Color);
end;

// Return the subproperty name.
function TRedElementProperty.GetName: string;
begin
  Result := sRed;
end;

// Return a new color, based on the old color, but with
// the red element changed to Value.
function TRedElementProperty.SetColorValue(Value: TColorElement; Color: TRGB): TRGB;
begin
  Result := RGB(Value, GetGValue(Color), GetBValue(Color));
end;

{ TGreenElement }
// Return the green part of a color.
function TGreenElementProperty.GetColorValue(Color: TRGB): TColorElement;
begin
  Result := GetGValue(Color);
end;

// Return the subproperty name.
function TGreenElementProperty.GetName: string;
begin
  Result := sGreen;
end;

// Return a new color, based on the old color, but with
// the green element changed to Value.
function TGreenElementProperty.SetColorValue(Value: TColorElement; Color: TRGB): TRGB;
begin
  Result := RGB(GetRValue(Color), Value, GetBValue(Color));
end;

{ TBlueElement }
// Return the blue part of a color.
function TBlueElementProperty.GetColorValue(Color: TRGB): TColorElement;
begin
  Result := GetBValue(Color);
end;

// Return the subproperty name.
function TBlueElementProperty.GetName: string;
begin
  Result := sBlue;
end;

// Return a new color, based on the old color, but with
// the blue element changed to Value.
function TBlueElementProperty.SetColorValue(Value: TColorElement; Color: TRGB): TRGB;
begin
  Result := RGB(GetRValue(Color), GetGValue(Color), Value);
end;

end.
