unit DateTimeProp;

// Date-time property editor that has subproperties for the
// elements of a date and time: year, month, day, hour, minute,
// second, and millisecond. This property editor demonstrates
// how you can use the TNestedProperty editor to create your
// own subproperties for any property.
//
// Copyright  1999 Tempest Software, Inc.

interface

uses Windows, SysUtils, Classes, Graphics, DsgnIntf;

type
  TDateTimeElement = type Word;

  // Parent property editor. Register TDataTimeSubproperty for any
  // property of type TDateTime. It's just like a TDateTimeProperty editor,
  // except that it has subproperties for the Year, Hour, etc.,
  // elements of a date and time. Do the same for TDateProperty
  // and TTimeProperty.
  TDateTimeSubproperty = class(TDateTimeProperty)
  public
    function GetAttributes: TPropertyAttributes; override;
    procedure GetProperties(Proc: TGetPropEditProc); override;
    procedure GetValues(Proc: TGetStrProc); override;
    procedure SetValue(const Value: string); override;
  end;
  TDateSubproperty = class(TDateProperty)
  public
    function GetAttributes: TPropertyAttributes; override;
    procedure GetProperties(Proc: TGetPropEditProc); override;
    procedure GetValues(Proc: TGetStrProc); override;
    procedure SetValue(const Value: string); override;
  end;
  TTimeSubproperty = class(TTimeProperty)
  public
    function GetAttributes: TPropertyAttributes; override;
    procedure GetProperties(Proc: TGetPropEditProc); override;
    procedure GetValues(Proc: TGetStrProc); override;
    procedure SetValue(const Value: string); override;
  end;

  TDateTimeElementKind = (dtYear, dtMonth, dtDay,
                          dtHours, dtMinutes, dtSeconds, dtMilliseconds);
  TDateTimeElements = array[TDateTimeElementKind] of Word;

  // Class for each date time element.
  TDateTimeElementProperty = class(TNestedProperty)
  private
    fKind: TDateTimeElementKind;
    fName: string;
  protected
    constructor Create(Parent: TPropertyEditor; Kind: TDateTimeElementKind; const Name: string);
    function GetElementValueAt(Index: Integer): TDateTimeElement;
    procedure SetElementValue(Value: TDateTimeElement);
    property Kind: TDateTimeElementKind read fKind;
  public
    function AllEqual: Boolean; override;
    procedure Edit; override;
    function GetName: string; override;
    function GetValue: string; override;
    procedure SetValue(const Value: string); override;
  end;
  // Months are special because the user can choose a month by name or number.
  TMonthProperty = class(TDateTimeElementProperty)
  public
    function GetAttributes: TPropertyAttributes; override;
    procedure GetValues(Proc: TGetStrProc); override;
    function GetValue: string; override;
    procedure SetValue(const Value: string); override;
  end;

procedure Register;

implementation

// Subproperty names.
resourcestring
  sYear = 'Year';
  sMonth = 'Month';
  sDay = 'Day';
  sHours = 'Hours';
  sMinutes = 'Minutes';
  sSeconds = 'Seconds';
  sMilliSeconds = 'MilliSeconds';
  sNow = 'Now';

// You need to register only the parent property editor. This editor
// replaces the standard TDateTimeProperty and related editors.
procedure Register;
begin
  RegisterPropertyEditor(TypeInfo(TDateTime), TPersistent, '', TDateTimeSubproperty);
  RegisterPropertyEditor(TypeInfo(TDateTime), TPersistent, 'Time', TTimeSubproperty);
  RegisterPropertyEditor(TypeInfo(TDateTime), TPersistent, 'Date', TDateSubproperty);
end;

{ TDateTimeSubproperty }
// TDateTimeSubproperty is just like TDateTimeProperty, except that it
// has subproperties.
function TDateTimeSubproperty.GetAttributes: TPropertyAttributes;
begin
  Result := inherited GetAttributes + [paSubproperties, paValueList];
end;

// Create each subproperty editor and pass it to Proc.
// Delphi frees the instances when it is done using them.
procedure TDateTimeSubproperty.GetProperties(Proc: TGetPropEditProc);
begin
  Proc(TDateTimeElementProperty.Create(Self, dtYear, sYear));
  Proc(TMonthProperty.Create(Self, dtMonth, sMonth));
  Proc(TDateTimeElementProperty.Create(Self, dtDay, sDay));
  Proc(TDateTimeElementProperty.Create(Self, dtHours, sHours));
  Proc(TDateTimeElementProperty.Create(Self, dtMinutes, sMinutes));
  Proc(TDateTimeElementProperty.Create(Self, dtSeconds, sSeconds));
  Proc(TDateTimeElementProperty.Create(Self, dtMilliSeconds, sMilliSeconds));
end;

procedure TDateTimeSubproperty.GetValues(Proc: TGetStrProc);
begin
  Proc(sNow);
end;

procedure TDateTimeSubproperty.SetValue(const Value: string);
begin
  if CompareText(Value, sNow) = 0 then
    SetFloatValue(Now)
  else
    inherited;
end;


{ TDateSubproperty }

function TDateSubproperty.GetAttributes: TPropertyAttributes;
begin
  Result := inherited GetAttributes + [paSubproperties, paValueList];
end;

procedure TDateSubproperty.GetProperties(Proc: TGetPropEditProc);
begin
  Proc(TDateTimeElementProperty.Create(Self, dtYear, sYear));
  Proc(TMonthProperty.Create(Self, dtMonth, sMonth));
  Proc(TDateTimeElementProperty.Create(Self, dtDay, sDay));
end;

procedure TDateSubproperty.GetValues(Proc: TGetStrProc);
begin
  Proc(sNow);
end;

procedure TDateSubproperty.SetValue(const Value: string);
begin
  if CompareText(Value, sNow) = 0 then
    SetFloatValue(Date)
  else
    inherited;
end;

{ TTimeSubproperty }

function TTimeSubproperty.GetAttributes: TPropertyAttributes;
begin
  Result := inherited GetAttributes + [paSubproperties, paValueList];
end;

procedure TTimeSubproperty.GetProperties(Proc: TGetPropEditProc);
begin
  Proc(TDateTimeElementProperty.Create(Self, dtHours, sHours));
  Proc(TDateTimeElementProperty.Create(Self, dtMinutes, sMinutes));
  Proc(TDateTimeElementProperty.Create(Self, dtSeconds, sSeconds));
  Proc(TDateTimeElementProperty.Create(Self, dtMilliSeconds, sMilliSeconds));
end;

procedure TTimeSubproperty.GetValues(Proc: TGetStrProc);
begin
  Proc(sNow);
end;

procedure TTimeSubproperty.SetValue(const Value: string);
begin
  if CompareText(Value, sNow) = 0 then
    SetFloatValue(Time)
  else
    inherited;
end;


{ TDateTimeElement }
constructor TDateTimeElementProperty.Create(Parent: TPropertyEditor;
  Kind: TDateTimeElementKind; const Name: string);
begin
  inherited Create(Parent);
  fKind := Kind;
  fName := Name;
end;

// Return True is all selected components have the same
// value for the color element.
function TDateTimeElementProperty.AllEqual: Boolean;
var
  I: Integer;
  Compare: TDateTimeElement;
begin
  if PropCount > 1 then
  begin
    // Get the color element of the first selected component.
    Compare := GetElementValueAt(0);
    for I := 1 to PropCount - 1 do
      // Compare the first with the rest of the components.
      if Compare <> GetElementValueAt(I) then
      begin
        // If any one is different, return False.
        Result := False;
        Exit;
      end;
  end;
  // They're all the same.
  Result := True;
end;

function MaxDayInMonth(Date: TDateTime): TDateTimeElement;
const
  MaxDays: array[Boolean, 1..12] of Byte =
   ({not leap year}(31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31),
        {leap year}(31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31));
var
  Year, Month, Day: Word;
begin
  DecodeDate(Date, Year, Month, Day);
  Result := MaxDays[IsLeapYear(Year), Month];
end;

// When the user double-clicks a value, increment it by one.
// Wrap around back to 1 when it reaches the limit.
procedure TDateTimeElementProperty.Edit;
var
  Value: TDateTimeElement;
begin
  Value := GetElementValueAt(0) + 1;
  case Kind of
  dtMonth:
    if Value > 12 then
      Value := 1;
  dtDay:
    if Value > MaxDayInMonth(GetFloatValue) then
      Value := 1;
  dtHours:
    if Value >= 24 then
      Value := 0;
  dtMinutes, dtSeconds:
    if Value >= 60 then
      Value := 0;
  dtMilliSeconds:
    if Value >= 1000 then
      Value := 0;
  else
    ; // let the value get as big as it wants to be
  end;
  SetElementValue(Value);
end;

// Return a string representation of the element value.
function TDateTimeElementProperty.GetValue: string;
begin
  Result := IntToStr(GetElementValueAt(0));
end;

// Set the parent date and time when the element changes to the
// given value. If "Value" is invalid, the Object Inspector
// will catch the exception that StrToInt raises.
procedure TDateTimeElementProperty.SetValue(const Value: string);
begin
  SetElementValue(StrToInt(Value));
end;

resourcestring
  sUnknownKind = 'Internal error: unknown date time element kind: %d';
  
function TDateTimeElementProperty.GetElementValueAt(
  Index: Integer): TDateTimeElement;
var
  Year, Month, Day, Hours, Minutes, Seconds, MilliSeconds: Word;
begin
  case Kind of
  dtYear:
    begin
      DecodeDate(GetFloatValueAt(Index), Year, Month, Day);
      Result := Year;
    end;
  dtMonth:
    begin
      DecodeDate(GetFloatValueAt(Index), Year, Month, Day);
      Result := Month;
    end;
  dtDay:
    begin
      DecodeDate(GetFloatValueAt(Index), Year, Month, Day);
      Result := Day;
    end;
  dtHours:
    begin
      DecodeTime(GetFloatValueAt(Index), Hours, Minutes, Seconds, Milliseconds);
      Result := Hours;
    end;
  dtMinutes:
    begin
      DecodeTime(GetFloatValueAt(Index), Hours, Minutes, Seconds, Milliseconds);
      Result := Minutes;
    end;
  dtSeconds:
    begin
      DecodeTime(GetFloatValueAt(Index), Hours, Minutes, Seconds, Milliseconds);
      Result := Seconds;
    end;
  dtMilliSeconds:
    begin
      DecodeTime(GetFloatValueAt(Index), Hours, Minutes, Seconds, Milliseconds);
      Result := MilliSeconds;
    end;
  else
    raise Exception.CreateFmt(sUnknownKind, [Ord(Kind)]);
  end;
end;

procedure TDateTimeElementProperty.SetElementValue(Value: TDateTimeElement);
var
  Year, Month, Day, Hours, Minutes, Seconds, MilliSeconds: Word;
begin
  DecodeDate(GetFloatValue, Year, Month, Day);
  DecodeTime(GetFloatValue, Hours, Minutes, Seconds, Milliseconds);
  case Kind of
  dtYear:
    Year := Value;
  dtMonth:
    Month := Value;
  dtDay:
    Day := Value;
  dtHours:
    Hours := Value;
  dtMinutes:
    Minutes := Value;
  dtSeconds:
    Seconds := Value;
  dtMilliSeconds:
    MilliSeconds := Value;
  else
    raise Exception.CreateFmt(sUnknownKind, [Ord(Kind)]);
  end;
  SetFloatValue(EncodeDate(Year, Month, Day) + EncodeTime(Hours, Minutes, Seconds, Milliseconds));
end;

function TDateTimeElementProperty.GetName: string;
begin
  Result := fName;
end;

{ TMonthProperty }

function TMonthProperty.GetAttributes: TPropertyAttributes;
begin
  Result := inherited GetAttributes + [paValueList];
end;

function TMonthProperty.GetValue: string;
begin
  Result := LongMonthNames[GetElementValueAt(0)];
end;

procedure TMonthProperty.GetValues(Proc: TGetStrProc);
var
  I: 1..12;
begin
  for I := Low(LongMonthNames) to High(LongMonthNames) do
    Proc(LongMonthNames[I]);
end;

procedure TMonthProperty.SetValue(const Value: string);
var
  I: 1..12;
begin
  for I := Low(LongMonthNames) to High(LongMonthNames) do
  begin
    if CompareText(LongMonthNames[I], Value) = 0 then
    begin
      SetElementValue(I);
      Exit;
    end
    else if CompareText(ShortMonthNames[I], Value) = 0 then
    begin
      SetElementValue(I);
      Exit;
    end
  end;
  // It's not a month name, so maybe it's a number.
  inherited SetValue(Value);
end;

end.
