unit SetProp;

interface

  { Implement an Edit dialog for a set property editor. The dialog box
    includes a check box for each set element, and the user can quickly
    check and uncheck elements. Select All and Select None buttons make
    it even easier to work with. The size of the dialog box is dynamically
    set, according to the widest check box label and the number of
    check boxes. One check box is created for each set element.

    Whether this edit dialog is actually easier to use than the default
    TSetProperty is up to the user to decide.
  }
uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, ExtCtrls, DsgnIntf, TypInfo, BoolProp, Menus;

type
  TSetEditForm = class;
  TSetEditProperty = class(TSetPropertyEx)
  private
    Form: TSetEditForm;
  public
    function GetAttributes: TPropertyAttributes; override;
    procedure Edit; override;
  end;
  TSetEditForm = class(TForm)
    CheckBoxPanel: TPanel;
    PopupMenu: TPopupMenu;
    SelectAll: TMenuItem;
    SelectNone: TMenuItem;
    N1: TMenuItem;
    AcceptItem: TMenuItem;
    CancelItem: TMenuItem;
    ResetItem: TMenuItem;
    procedure FormDeactivate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormDestroy(Sender: TObject);
    procedure CancelItemClick(Sender: TObject);
    procedure ResetItemClick(Sender: TObject);
    procedure AcceptItemClick(Sender: TObject);
    procedure SelectAllClick(Sender: TObject);
    procedure SelectNoneClick(Sender: TObject);
    procedure CheckBoxClick(Sender: TObject);
  private
    { Private declarations }
    fEditor: TSetEditProperty;    { keep a pointer to the property editor }
    fTypeData: PTypeData;         { cache the type data, for convenience }
    fResetValue: Integer;
    procedure SetupForm;          { create the check boxes, resize form, etc. }
    procedure SetSetValue(Value: LongInt);
    function GetSetValue: LongInt;
    function GetMinValue: Integer;
    function GetMaxValue: Integer;
    function GetEnumType: PTypeInfo;
    function GetCheckBox(Index: Integer): TCheckBox;
  protected
    { convenience properties, to return type information }
    property MinValue: Integer read GetMinValue;
    property MaxValue: Integer read GetMaxValue;
    property TypeData: PTypeData read fTypeData;
    property EnumType: PTypeInfo read GetEnumType;
    property ResetValue: Integer read fResetValue;
  public
    { Public declarations }
    constructor Create(Owner: TComponent; Editor: TSetEditProperty); reintroduce;

    { Return the property editor }
    property Editor: TSetEditProperty read fEditor;
    { Set or Get the actual value of the set, as an ordinal value }
    property Value: LongInt read GetSetValue write SetSetValue;
    { Return a check box, indexed by set element }
    property CheckBox[Index: Integer]: TCheckBox read GetCheckBox;
  end;

procedure Register;

implementation

uses Grids;

{$R *.DFM}

procedure Register;
begin
  // Unfortunately, you must register TSetEditProperty for each
  // distinct set type. Here are some examples.
  RegisterPropertyEditor(TypeInfo(TFontStyles), nil, '', TSetEditProperty);
  RegisterPropertyEditor(TypeInfo(TBorderIcons), nil, '', TSetEditProperty);
  RegisterPropertyEditor(TypeInfo(TMenuAnimation), nil, '', TSetEditProperty);
  RegisterPropertyEditor(TypeInfo(TAnchors), nil, '', TSetEditProperty);
  RegisterPropertyEditor(TypeInfo(TGridOptions), nil, '', TSetEditProperty);
  RegisterPropertyEditor(TypeInfo(TOpenOptions), nil, '', TSetEditProperty);
  RegisterPropertyEditor(TypeInfo(TFontDialogOptions), nil, '', TSetEditProperty);
  RegisterPropertyEditor(TypeInfo(TColorDialogOptions), nil, '', TSetEditProperty);
  RegisterPropertyEditor(TypeInfo(TPrintDialogOptions), nil, '', TSetEditProperty);
  RegisterPropertyEditor(TypeInfo(TFindOptions), nil, '', TSetEditProperty);
end;

{ TSetEditProperty }

procedure TSetEditProperty.Edit;
begin
  if Form = nil then
    Form := TSetEditForm.Create(Application, Self);
  Form.Value := GetOrdValue;
  Form.Show;
end;

function TSetEditProperty.GetAttributes: TPropertyAttributes;
begin
  Result := inherited GetAttributes + [paDialog] - [paValueList];
end;

{ TSetEditForm }

{ Make it easy to refer to a set element with an integer index }
const
  BitsPerByte = 8;
type
  TIntegerSet = set of 0..SizeOf(Integer)*BitsPerByte - 1;

{ Create the form and set up the check boxes, etc. }
constructor TSetEditForm.Create(Owner: TComponent; Editor: TSetEditProperty);
begin
  inherited Create(Owner);
  fEditor := Editor;
  fTypeData := GetTypeData(fEditor.GetPropType);
  fResetValue := Editor.GetOrdValue;
  SetupForm;
end;

{ Return the type info for the set component type, which is an enumerated type }
function TSetEditForm.GetEnumType: PTypeInfo;
begin
  Result := TypeData^.CompType^
end;

{ Return the minimum ordinal index for the set elements (always 0) }
function TSetEditForm.GetMinValue: Integer;
begin
  with GetTypeData(EnumType)^ do
    Result := MinValue
end;

{ Return the maximum ordinal index for the set elements (# set elements-1) }
function TSetEditForm.GetMaxValue: Integer;
begin
  with GetTypeData(EnumType)^ do
    Result := MaxValue
end;

{ Return a check box, indexed by set element index. Since all
  the check boxes are child controls of the CheckBoxPanel
  component, it is easy to get the right one.  The only children
  of the CheckBoxPanel are the check boxes. }
function TSetEditForm.GetCheckBox(Index: Integer): TCheckBox;
begin
  Result := CheckBoxPanel.Controls[Index-MinValue] as TCheckBox;
end;

{ Return the current value of the set, according to which boxes are checked. }
function TSetEditForm.GetSetValue: LongInt;
var
  I: Integer;
  Value: TIntegerSet;
begin
  Value := [];
  for I := MinValue to MaxValue do
    if CheckBox[I].Checked then
      Include(Value, I);
  Result := Integer(Value);
end;

{ Set the check boxes to reflect the given set value. In Delphi 1.0,
  the value is in the range 0..65535, but in Delphi32, the range is
  the full integer range. Thus, when we cast the ordinal value to
  the TIntegerSet, we need to use a different cast in Delphi 1.0
  and Delphi 32. }
procedure TSetEditForm.SetSetValue(Value: LongInt);
var
  I: Integer;
  SetValue: TIntegerSet;
begin
  SetValue := TIntegerSet(Value);
  for I := MinValue to MaxValue do
    CheckBox[I].Checked := I in SetValue;
end;

{ Set up the form with all the check boxes. }
procedure TSetEditForm.SetupForm;
const
  MaxBoxesPerColumn = 16;        { Must be power of 2 }
  Margin = 8;                    { Margin around visible components }
  BigMargin = 2*Margin;
  VertMargin = 2;                { Vertical space between check boxes }
type
  TNumColumns = 1..SizeOf(Integer)*BitsPerByte div MaxBoxesPerColumn;
var
  I: Integer;
  TmpWidth: Integer;
  Column: TNumColumns;
  MaxWidth: array[TNumColumns] of Integer; { check box width for each column }
  X, Y: Integer;                           { check box position }
  MaxY: Integer;                           { Max Y for any column }
  CheckBox: TCheckBox;                     { check box being created }
  Pos: TPoint;
  Rect: TRect;
  CheckBoxHeight: Integer;
begin
  // First determine the widest check box label in each column.
  for Column := Low(TNumColumns) to High(TNumColumns) do
    MaxWidth[Column] := 0;
  Column := Low(TNumColumns);
  for I := MinValue to MaxValue do
  begin
    if (I>MinValue) and ((I-MinValue) mod MaxBoxesPerColumn = 0) then
      Column := Column + 1;
    TmpWidth := Canvas.TextWidth(GetEnumName(EnumType, I));
    if TmpWidth > MaxWidth[Column] then
      MaxWidth[Column] := TmpWidth;
  end;
  
  // Allow room for the check box itself.
  for Column := Low(TNumColumns) to High(TNumColumns) do
    MaxWidth[Column] := MaxWidth[Column] + 20;

  { Create the check boxes, positioning them one below the other.
    Set the label of each check box to the enumerated literal name.
    Put all the check boxes on the CheckBoxPanel form. See the GetCheckBox
    function to learn why. }
  X := Margin;
  Y := Margin;
  MaxY := Y;
  Column := Low(TNumColumns);
  CheckBoxHeight := 0;
  for I := MinValue to MaxValue do
  begin
    CheckBox := TCheckBox.Create(Self);
    CheckBoxHeight := CheckBox.Height;
    if (I>MinValue) and ((I-MinValue) mod MaxBoxesPerColumn = 0) then
    begin
      Y := Margin;
      X := Margin + CheckBoxHeight + MaxWidth[Column];
      Column := Column + 1;
    end;
    CheckBox.OnClick := CheckBoxClick;
    CheckBox.Tag := I;
    CheckBox.Left := X;
    CheckBox.Top  := Y;
    CheckBox.Width := MaxWidth[Column];
    CheckBox.Caption := GetEnumName(EnumType, I);
    CheckBox.Parent := CheckBoxPanel;
    Y := Y + CheckBoxHeight + VertMargin;
    if Y > MaxY then
      MaxY := Y;
  end;

  // Locate the control that surrounds the cursor,
  // to position the form next to the control.
  GetCursorPos(Pos);
  if not GetWindowRect(WindowFromPoint(Pos), Rect) then
  begin
    Rect.TopLeft := Pos;
    Rect.BottomRight := Pos;
  end;

  // Resize the form to accommodate all the check boxes.
  TmpWidth := 0;
  for Column := Low(TNumColumns) to High(TNumColumns) do
    TmpWidth := TmpWidth + MaxWidth[Column] + CheckBoxHeight;
  // Move the form's origin to just below or just above the
  // control where the user clicked
  if Rect.Bottom + Y > Screen.Height then
    Rect.Top := Rect.Top - Y
  else
    Rect.Top := Rect.Bottom;
  if Rect.Right + TmpWidth > Screen.Width then
    Rect.Left := Rect.Left - TmpWidth;
  SetBounds(Rect.Left, Rect.Top, TmpWidth, MaxY);

  // Make the first check box the active control, so the user can easily
  // start checking and unchecking boxes.
  ActiveControl := CheckBoxPanel.Controls[0] as TWinControl;
end;

procedure TSetEditForm.FormDeactivate(Sender: TObject);
begin
  Editor.SetOrdValue(Value);
  Release;
end;

procedure TSetEditForm.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  Action := caFree;
end;

procedure TSetEditForm.FormDestroy(Sender: TObject);
begin
  if Editor.Form = Self then
    Editor.Form := nil;
end;

procedure TSetEditForm.CancelItemClick(Sender: TObject);
begin
  Editor.SetOrdValue(ResetValue);
  Release;
end;

procedure TSetEditForm.ResetItemClick(Sender: TObject);
begin
  Value := ResetValue;
  Editor.SetOrdValue(Value);
end;

procedure TSetEditForm.AcceptItemClick(Sender: TObject);
begin
  Release;
end;

procedure TSetEditForm.SelectAllClick(Sender: TObject);
var
  I: Integer;
begin
  for I := MinValue to MaxValue do
    CheckBox[I].Checked := True;
end;

procedure TSetEditForm.SelectNoneClick(Sender: TObject);
var
  I: Integer;
begin
  for I := MinValue to MaxValue do
    CheckBox[I].Checked := False;
end;

procedure TSetEditForm.CheckBoxClick(Sender: TObject);
begin
  Editor.SetOrdValue(Value);
end;

end.
