unit VisualPropEds;

// Visual property editors.
// Delphi 5 introduces owner-drawn property editors. These property editors
// demonstrate some of the owner-drawn features.
//
// TVisualFontProperty displays the selected font name and size, using
// the chosen font name and style (but not the size).
//
// TVisualGraphicProperty displays the selected graphic, scaled to fit
// in the object inspector.
//
// TVisualPictureProperty displays the selected picture, scaled to fit
// in the object inspector.
//
// Copyright  1999 Tempest Software, Inc.

interface

uses Windows, SysUtils, Classes, Graphics, DsgnIntf, PicEdit;

type
  TVisualFontProperty = class(TFontProperty)
  public
    function GetValue: string; override;
    procedure PropDrawValue(Canvas: TCanvas; const Rect: TRect; Selected: Boolean); override;
  end;

  TVisualGraphicProperty = class(TGraphicProperty)
  public
    procedure PropDrawValue(Canvas: TCanvas; const Rect: TRect; Selected: Boolean); override;
  end;

  TVisualPictureProperty = class(TPictureProperty)
  public
    procedure PropDrawValue(Canvas: TCanvas; const Rect: TRect; Selected: Boolean); override;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterPropertyEditor(TypeInfo(TFont), nil, '', TVisualFontProperty);
  RegisterPropertyEditor(TypeInfo(TGraphic), nil, '', TVisualGraphicProperty);
  RegisterPropertyEditor(TypeInfo(TPicture), nil, '', TVisualPictureProperty);
end;

{ TVisualFontProperty }

function TVisualFontProperty.GetValue: string;
var
  Font: TFont;
begin
  Font := TFont(GetOrdValue);
  if Font = nil then
    Result := inherited GetValue
  else
    Result := Format('%s, %d', [Font.Name, Font.Size]);
end;

procedure TVisualFontProperty.PropDrawValue(Canvas: TCanvas;
  const Rect: TRect; Selected: Boolean);
var
  Font: TFont;
begin
  // Instead of a boring (TFont) font value, show the font name and size
  // using the chosen font style. The user might have chosen a large size,
  // so select only the font name, leaving the size as the default size.
  // Set the font color only if it is different from the background color,
  // or else the user could not see the font name.
  Font := TFont(GetOrdValue);
  if Font <> nil then
  begin
    if ColorToRGB(Font.Color) <> ColorToRGB(clBtnFace) then
      Canvas.Font.Color := Font.Color;
    Canvas.Font.Name := Font.Name;
    Canvas.Font.Style := Font.Style;
  end;
  inherited;
end;

{ TVisualGraphicProperty }

procedure StretchIcon(Canvas: TCanvas; const Rect: TRect; Icon: TIcon);
var
  Bitmap: TBitmap;
begin
  // Windows does not stretch icons, so if the icon doesn't fit,
  // draw it on a temporary bitmap and stretch the bitmap.
  Bitmap := TBitmap.Create;
  try
    Bitmap.Height := Icon.Height;
    Bitmap.Width  := Icon.Width;
    Bitmap.Canvas.Brush.Color := clBtnFace;
    Bitmap.Canvas.FillRect(Rect);
    Bitmap.Canvas.Draw(0, 0, Icon);
    Canvas.StretchDraw(Rect, Bitmap);
  finally
    Bitmap.Free;
  end;
end;

procedure DrawGraphic(Canvas: TCanvas; const Rect: TRect;
                      Graphic: TGraphic; const Value: string);
var
  R: TRect;
  HeightRatio, WidthRatio: Single;
begin
  Canvas.FillRect(Rect);
  // Fit the graphic into the given space. Maintain the aspect ratio
  // and adjust the height or width to fill the given space.
  HeightRatio := (Rect.Bottom - Rect.Top) / Graphic.Height;
  WidthRatio  := (Rect.Right - Rect.Left) / Graphic.Width;
  R := Rect;
  if HeightRatio < WidthRatio then
    R.Right := R.Left + Trunc(Graphic.Width * HeightRatio)
  else
    R.Bottom := R.Top + Trunc(Graphic.Height * WidthRatio);
  // Windows does not stretch icons, so if the icon doesn't fit,
  // draw it on a temporary bitmap and stretch the bitmap.
  if (Graphic is TIcon) and ((HeightRatio > 1) or (WidthRatio > 1)) then
    StretchIcon(Canvas, R, TIcon(Graphic))
  else
    Canvas.StretchDraw(R, Graphic);

  // To the right of the graphic, let the inherited editor draw the
  // the usual text, e.g., (TIcon).
  R.Left := R.Right;
  R.Right := Rect.Right;
  R.Top := Rect.Top;
  R.Bottom := Rect.Bottom;
  Canvas.TextRect(R, R.Left+1, R.Top+1, Value);
end;

procedure TVisualGraphicProperty.PropDrawValue(Canvas: TCanvas;
  const Rect: TRect; Selected: Boolean);
var
  Graphic: TGraphic;
begin
  Graphic := TGraphic(GetOrdValue);
  if (Graphic = nil) or Graphic.Empty or (Graphic.Height = 0) or (Graphic.Width = 0) then
    inherited
  else
    DrawGraphic(Canvas, Rect, Graphic, GetVisualValue);
end;


{ TVisualPictureProperty }

procedure TVisualPictureProperty.PropDrawValue(Canvas: TCanvas;
  const Rect: TRect; Selected: Boolean);
var
  Graphic: TGraphic;
begin
  if GetOrdValue = 0 then
    inherited
  else
  begin
    Graphic := TPicture(GetOrdValue).Graphic;
    if (Graphic = nil) or Graphic.Empty or (Graphic.Height = 0) or (Graphic.Width = 0) then
      inherited
    else
      DrawGraphic(Canvas, Rect, Graphic, GetVisualValue);
  end;
end;

end.
